import os
import re
import subprocess
import sys
import unittest


class TestPackageBasics(unittest.TestCase):
    def test_version_string(self):
        import lrcgen

        self.assertIsInstance(lrcgen.__version__, str)
        self.assertRegex(lrcgen.__version__, r"^\d+\.\d+\.\d+$")

    def test_api_import_is_light(self):
        # Should import without forcing whisper/openai at import-time.
        from lrcgen.api import generate_lrc_sync, generate_lrc_batch_sync

        self.assertTrue(callable(generate_lrc_sync))
        self.assertTrue(callable(generate_lrc_batch_sync))


class TestCliEntry(unittest.TestCase):
    def _run(self, args):
        return subprocess.run(
            [sys.executable, "-m", "lrcgen", *args],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=False,
        )

    def test_help(self):
        p = self._run(["--help"])
        self.assertEqual(p.returncode, 0)
        self.assertIn("usage:", p.stdout.lower())

    def test_version(self):
        p = self._run(["--version"])
        self.assertEqual(p.returncode, 0)
        self.assertRegex(p.stdout.strip(), r"^lrcgen\s+\d+\.\d+\.\d+$")


class TestCorrectorInternals(unittest.TestCase):
    def test_parse_numbered_lines(self):
        from lrcgen.corrector import _parse_numbered_lines

        text = """
1: 第一行
2: 第二行
3: 第三行
""".strip()
        out = _parse_numbered_lines(text, 3)
        self.assertEqual(out, ["第一行", "第二行", "第三行"])

    def test_guardrail_merge_rejects_big_changes(self):
        from lrcgen.corrector import _guardrail_merge

        original = ["年过花 热了白", "月下举杯 尽沧海"]
        candidate = ["这是完全不同的一句话", "月下举杯 尽沧海"]
        merged = _guardrail_merge(original, candidate)
        self.assertEqual(merged[0], original[0])
        self.assertEqual(merged[1], original[1])

    def test_guardrail_merge_accepts_small_fix(self):
        from lrcgen.corrector import _guardrail_merge

        original = ["年过花 热了白"]
        candidate = ["年过花 落了白"]
        merged = _guardrail_merge(original, candidate)
        # Should allow a single-character correction by default.
        self.assertEqual(merged, candidate)


if __name__ == "__main__":
    unittest.main()


from __future__ import annotations

import re
import asyncio
import sys
import os
from difflib import SequenceMatcher

try:
    from openai import OpenAI
except ModuleNotFoundError:  # optional dependency
    OpenAI = None
from .config import (
    OPENAI_API_KEY,
    OPENAI_BASE_URL,
    OPENAI_MODEL,
    MAX_LLM_CONCURRENCY,
)

# ===== OpenAI Client =====
client = None
if OpenAI and OPENAI_API_KEY:
    client = OpenAI(
        api_key=OPENAI_API_KEY,
        base_url=OPENAI_BASE_URL
    )

_semaphore = asyncio.Semaphore(MAX_LLM_CONCURRENCY)


def _similarity(a: str, b: str) -> float:
    # 0..1, higher means more similar
    if a == b:
        return 1.0
    if not a or not b:
        return 0.0
    return SequenceMatcher(None, a, b).ratio()


def _guardrail_merge(original: list[str], candidate: list[str]) -> list[str]:
    """Keep candidate only when it's a small, likely-safe correction.

    This prevents the LLM from paraphrasing/reformatting lines (which often
    makes lyrics worse). Lines that change too much are reverted to original.
    """

    min_sim = float(os.getenv("LRCGEN_LLM_MIN_SIM", "0.78"))
    max_len_delta = int(os.getenv("LRCGEN_LLM_MAX_LEN_DELTA", "6"))

    out: list[str] = []
    for o, c in zip(original, candidate):
        o2 = o.strip()
        c2 = c.strip()

        # Never allow empty replacements
        if not c2:
            out.append(o)
            continue

        # If the model changed the line too much, keep the original
        if abs(len(o2) - len(c2)) > max_len_delta:
            out.append(o)
            continue

        if _similarity(o2, c2) < min_sim:
            out.append(o)
            continue

        out.append(c2)

    return out


def _guardrail_merge_with_stats(
    original: list[str],
    candidate: list[str],
) -> tuple[list[str], int, int, int]:
    """Return (merged_lines, proposed_changes, accepted_changes, rejected_changes)."""

    proposed_changes = sum(
        1 for o, c in zip(original, candidate) if o.strip() != c.strip()
    )

    merged = _guardrail_merge(original, candidate)
    accepted_changes = sum(
        1 for o, m in zip(original, merged) if o.strip() != m.strip()
    )
    rejected_changes = max(0, proposed_changes - accepted_changes)
    return merged, proposed_changes, accepted_changes, rejected_changes


_LINE_RE = re.compile(r"^\s*(\d+)\s*[\.|:：\)]\s*(.*)$")


def _parse_numbered_lines(text: str, n: int) -> list[str] | None:
    """Parse lines in format like '1: ...' .. 'n: ...' (order can vary)."""

    mapping: dict[int, str] = {}
    for raw in text.splitlines():
        m = _LINE_RE.match(raw)
        if not m:
            continue
        idx = int(m.group(1))
        if 1 <= idx <= n and idx not in mapping:
            mapping[idx] = m.group(2).strip()

    if len(mapping) != n:
        return None

    return [mapping[i] for i in range(1, n + 1)]


def needs_llm_fix(lines: list[str]) -> bool:
    """
    判断是否值得花钱纠错
    """
    if len(lines) < 8:
        return False

    total = len(lines)
    short = sum(1 for l in lines if len(l) <= 2)
    trad = sum(1 for l in lines if re.search(r"[裏爲妳祢著]", l))

    if short / total > 0.3:
        return True
    if trad / total > 0.2:
        return True

    return False


async def llm_fix(lines: list[str], *, require: bool = False) -> list[str]:
    """
    二次纠正（失败自动回退）
    """
    if not OpenAI:
        if require:
            raise RuntimeError("missing dependency 'openai'. Install with: pip install openai")
        return lines

    if not client:
        if require:
            raise RuntimeError("OPENAI_API_KEY not set")
        return lines

    async with _semaphore:
        try:
            numbered = "\n".join(f"{i+1}: {l}" for i, l in enumerate(lines))
            prompt = f"""
你是一个非常保守的中文歌词纠错器。目标：只做“确定无疑”的错别字/同音字修正。

硬性规则（必须严格遵守）：
1) 修正明显错别字、同音字（例如：热了白→落了白）；
2) 不允许改写句子、调整语序、增删标点/空格风格、替换近义词
3) 必须逐行对齐：输入有 {len(lines)} 行，你也必须输出 {len(lines)} 行
4) 每行必须带行号，格式必须是："<行号>: <内容>"，行号从 1 开始
5) 除了这 {len(lines)} 行之外，不要输出任何解释、标题、空行
6) 大多给你的都是已有的流行歌曲，请先检索出歌曲正确歌词，尽量参考正确歌词拼写

输入歌词：
{numbered}
""".strip()

            resp = client.chat.completions.create(
                model=OPENAI_MODEL,
                messages=[{"role": "user", "content": prompt}],
                temperature=0.1
            )

            content = (resp.choices[0].message.content or "").strip()
            parsed = _parse_numbered_lines(content, len(lines))
            if not parsed:
                print(
                    "[lrcgen] LLM correction ignored: failed to parse numbered output",
                    file=sys.stderr,
                )
                return lines

            # Guardrails: keep only safe/small corrections per line
            merged, proposed, accepted, rejected = _guardrail_merge_with_stats(
                lines, parsed
            )
            print(
                f"[lrcgen] LLM changes: proposed={proposed}, accepted={accepted}, rejected={rejected}",
                file=sys.stderr,
            )
            return merged

        except Exception as e:
            print(f"[lrcgen] LLM correction failed: {e}", file=sys.stderr)
            return lines



import re

try:
    from opencc import OpenCC

    _cc = OpenCC("t2s")
except ModuleNotFoundError:  # optional dependency
    _cc = None


def format_time(seconds: float) -> str:
    minutes = int(seconds // 60)
    seconds = seconds % 60
    return f"[{minutes:02d}:{seconds:05.2f}]"


def basic_clean(text: str) -> str:
    """
    繁转简 + 基础清洗
    """
    if _cc:
        text = _cc.convert(text)
    text = re.sub(r"[，。！？、；：]", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import List, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from .whisperer import WhisperRecognizer

AUDIO_EXTS: Tuple[str, ...] = (".mp3", ".wav", ".flac", ".m4a", ".ogg")


@dataclass(frozen=True)
class GenerateResult:
    audio_path: str
    out_path: str
    used_llm: bool
    line_count: int


async def generate_lrc(
    audio_path: str,
    out_path: str,
    *,
    mode: str = "auto",
    recognizer: Optional["WhisperRecognizer"] = None,
    model_name: str = "medium",
    language: str = "zh",
) -> GenerateResult:
    """Generate a single .lrc file.

    mode: "offline" | "online" | "auto"
    """

    if mode not in {"offline", "online", "auto"}:
        raise ValueError(f"invalid mode: {mode}")

    if recognizer is None:
        try:
            from .whisperer import WhisperRecognizer
        except ModuleNotFoundError as e:
            if e.name == "whisper":
                raise RuntimeError(
                    "missing dependency 'openai-whisper'. Install with: pip install openai-whisper"
                )
            raise
        recognizer = WhisperRecognizer(model_name=model_name, language=language)

    times, lines = recognizer.transcribe(audio_path)

    used_llm = False
    if mode == "online":
        used_llm = True
        from .corrector import llm_fix

        lines = await llm_fix(lines, require=True)
    elif mode == "auto":
        from .corrector import needs_llm_fix

        if not needs_llm_fix(lines):
            used_llm = False
        else:
            used_llm = True
            from .corrector import llm_fix

            lines = await llm_fix(lines, require=False)

    title = os.path.splitext(os.path.basename(audio_path))[0]

    from .utils import format_time

    with open(out_path, "w", encoding="utf-8") as f:
        f.write(f"[ti:{title}]\n[ar:Unknown]\n[al:Unknown]\n\n")
        for t, l in zip(times, lines):
            f.write(f"{format_time(t)}{l}\n")

    return GenerateResult(
        audio_path=audio_path,
        out_path=out_path,
        used_llm=used_llm,
        line_count=len(lines),
    )


async def generate_lrc_batch(
    input_dir: str,
    output_dir: str,
    *,
    mode: str = "auto",
    recognizer: Optional["WhisperRecognizer"] = None,
    model_name: str = "medium",
    language: str = "zh",
) -> List[GenerateResult]:
    """Generate .lrc files for all supported audio files under input_dir."""

    if mode not in {"offline", "online", "auto"}:
        raise ValueError(f"invalid mode: {mode}")

    os.makedirs(output_dir, exist_ok=True)
    if recognizer is None:
        try:
            from .whisperer import WhisperRecognizer
        except ModuleNotFoundError as e:
            if e.name == "whisper":
                raise RuntimeError(
                    "missing dependency 'openai-whisper'. Install with: pip install openai-whisper"
                )
            raise
        recognizer = WhisperRecognizer(model_name=model_name, language=language)

    results: List[GenerateResult] = []
    for name in os.listdir(input_dir):
        if not name.lower().endswith(AUDIO_EXTS):
            continue
        in_path = os.path.join(input_dir, name)
        out_path = os.path.join(output_dir, os.path.splitext(name)[0] + ".lrc")
        results.append(
            await generate_lrc(
                in_path,
                out_path,
                mode=mode,
                recognizer=recognizer,
                model_name=model_name,
                language=language,
            )
        )

    return results


def generate_lrc_sync(
    audio_path: str,
    out_path: str,
    *,
    mode: str = "auto",
    model_name: str = "medium",
    language: str = "zh",
) -> GenerateResult:
    import asyncio

    return asyncio.run(
        generate_lrc(
            audio_path,
            out_path,
            mode=mode,
            recognizer=None,
            model_name=model_name,
            language=language,
        )
    )


def generate_lrc_batch_sync(
    input_dir: str,
    output_dir: str,
    *,
    mode: str = "auto",
    model_name: str = "medium",
    language: str = "zh",
) -> List[GenerateResult]:
    import asyncio

    return asyncio.run(
        generate_lrc_batch(
            input_dir,
            output_dir,
            mode=mode,
            recognizer=None,
            model_name=model_name,
            language=language,
        )
    )



import argparse
import asyncio
import os
import sys

AUDIO_EXTS = (".mp3", ".wav", ".flac", ".m4a", ".ogg")


def main():
    parser = argparse.ArgumentParser("lrcgen")
    from . import __version__

    parser.add_argument("--version", action="version", version=f"lrcgen {__version__}")
    # Back-compat positional args (older usage: lrcgen input_dir output_dir)
    parser.add_argument("input", nargs="?", help="音频文件或目录（兼容旧版：目录）")
    parser.add_argument("output", nargs="?", help="输出文件或目录（兼容旧版：目录）")
    # Flag-style args (new usage: --audio <file/dir> --output <file/dir>)
    parser.add_argument("--audio", dest="audio", help="音频文件或目录")
    parser.add_argument("--output", dest="out", help="输出文件或目录")
    parser.add_argument("--offline", action="store_true", help="完全离线")
    parser.add_argument("--online", action="store_true", help="强制联网纠错")
    parser.add_argument("--model", default="medium", help="Whisper model (e.g. tiny/base/small/medium/large)")
    parser.add_argument("--language", default="zh", help="Whisper language code (default: zh)")
    args = parser.parse_args()

    input_path = args.audio or args.input
    output_path = args.out or args.output

    if not input_path or not output_path:
        parser.error("missing input/output: use positional 'input output' or flags '--audio/--output'")

    if not os.path.exists(input_path):
        parser.error(f"input not found: {input_path}")

    if args.offline:
        mode = "offline"
    elif args.online:
        mode = "online"
    else:
        mode = "auto"

    if mode == "online":
        from .config import OPENAI_API_KEY

        try:
            import openai  # noqa: F401
        except ModuleNotFoundError:
            parser.error("--online requires dependency 'openai'. Install with: pip install openai")

        if not OPENAI_API_KEY:
            parser.error(
                "--online requires OPENAI_API_KEY. Set it in .env or environment variables."
            )

    is_input_dir = os.path.isdir(input_path)
    if is_input_dir:
        os.makedirs(output_path, exist_ok=True)
    else:
        out_parent = os.path.dirname(output_path)
        if out_parent:
            os.makedirs(out_parent, exist_ok=True)

    try:
        from .whisperer import WhisperRecognizer
    except ModuleNotFoundError as e:
        if e.name == "whisper":
            parser.error(
                "missing dependency 'openai-whisper'. Install with: pip install openai-whisper"
            )
        raise

    recognizer = WhisperRecognizer(model_name=args.model, language=args.language)

    from .api import generate_lrc, generate_lrc_batch

    async def runner():
        if is_input_dir:
            results = await generate_lrc_batch(
                input_path,
                output_path,
                mode=mode,
                recognizer=recognizer,
            )
            for r in results:
                if r.used_llm:
                    print(
                        f"[lrcgen] LLM corrected: {os.path.basename(r.audio_path)}",
                        file=sys.stderr,
                    )
                print(f"[lrcgen] wrote: {r.out_path}", file=sys.stderr)
            return

        # Single file mode
        if not input_path.lower().endswith(AUDIO_EXTS):
            parser.error(f"unsupported audio extension: {input_path}")

        # If output is a directory, write <basename>.lrc into it
        if os.path.isdir(output_path) or output_path.endswith(os.sep):
            os.makedirs(output_path, exist_ok=True)
            out_path = os.path.join(
                output_path,
                os.path.splitext(os.path.basename(input_path))[0] + ".lrc",
            )
        else:
            out_path = output_path
            if not out_path.lower().endswith(".lrc"):
                out_path = out_path + ".lrc"

        r = await generate_lrc(
            input_path,
            out_path,
            mode=mode,
            recognizer=recognizer,
            model_name=args.model,
            language=args.language,
        )
        if r.used_llm:
            print(f"[lrcgen] LLM corrected: {os.path.basename(r.audio_path)}", file=sys.stderr)
        print(f"[lrcgen] wrote: {r.out_path}", file=sys.stderr)

    asyncio.run(runner())


if __name__ == "__main__":
    main()



import whisper
from .utils import basic_clean


class WhisperRecognizer:
    def __init__(self, model_name="medium", language="zh"):
        self.model = whisper.load_model(model_name)
        self.language = language

    def transcribe(self, audio_path: str):
        """
        返回：
        times: [float]
        lines: [str]
        """
        result = self.model.transcribe(
            audio_path,
            language=self.language,
            verbose=False
        )

        times, lines = [], []
        for seg in result["segments"]:
            text = basic_clean(seg["text"])
            if not text:
                continue
            times.append(seg["start"])
            lines.append(text)

        return times, lines



import os
from dotenv import load_dotenv

from dataclasses import dataclass
from typing import Optional

load_dotenv()

# ===== OpenAI / 兼容 API =====
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
OPENAI_BASE_URL = os.getenv("OPENAI_BASE_URL", "https://api.openai.com/v1")
OPENAI_MODEL = os.getenv("OPENAI_MODEL", "gpt-4o-mini")

# ===== LLM 并发限制 =====
MAX_LLM_CONCURRENCY = int(os.getenv("MAX_LLM_CONCURRENCY", "2"))


@dataclass(frozen=True)
class Settings:
    OPENAI_API_KEY: Optional[str] = OPENAI_API_KEY
    OPENAI_BASE_URL: str = OPENAI_BASE_URL
    OPENAI_MODEL: str = OPENAI_MODEL
    MAX_LLM_CONCURRENCY: int = MAX_LLM_CONCURRENCY


# Convenience accessor (for debugging / external scripts)
settings = Settings()
